/*
 * Copyright (c) 2017 SpaceToad and the BuildCraft team
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/
 */

package buildcraft.builders.snapshot;

import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.UUID;
import java.util.function.Predicate;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import javax.annotation.Nonnull;

import org.apache.commons.lang3.tuple.Pair;

import net.minecraft.entity.Entity;
import net.minecraft.entity.EntityHanging;
import net.minecraft.entity.EntityList;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTBase;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.nbt.NBTUtil;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.ResourceLocation;
import net.minecraft.util.Rotation;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Vec3d;
import net.minecraft.world.World;

import net.minecraftforge.fluids.FluidStack;

import buildcraft.api.core.InvalidInputDataException;
import buildcraft.api.schematics.ISchematicEntity;
import buildcraft.api.schematics.SchematicEntityContext;

import buildcraft.lib.misc.NBTUtilBC;
import buildcraft.lib.misc.RotationUtil;

public class SchematicEntityDefault implements ISchematicEntity {
    private NBTTagCompound entityNbt;
    private Vec3d pos;
    private BlockPos hangingPos;
    private EnumFacing hangingFacing;
    private Rotation entityRotation = Rotation.NONE;

    public static boolean predicate(SchematicEntityContext context) {
        ResourceLocation registryName = EntityList.func_191301_a(context.entity);
        return registryName != null &&
            RulesLoader.READ_DOMAINS.contains(registryName.func_110624_b()) &&
            RulesLoader.getRules(
                EntityList.func_191301_a(context.entity),
                context.entity.serializeNBT()
            )
                .stream()
                .anyMatch(rule -> rule.capture);
    }

    @Override
    public void init(SchematicEntityContext context) {
        entityNbt = context.entity.serializeNBT();
        pos = context.entity.func_174791_d().func_178788_d(new Vec3d(context.basePos));
        if (context.entity instanceof EntityHanging) {
            EntityHanging entityHanging = (EntityHanging) context.entity;
            hangingPos = entityHanging.func_174857_n().func_177973_b(context.basePos);
            hangingFacing = entityHanging.func_174811_aO();
        } else {
            hangingPos = new BlockPos(pos);
            hangingFacing = EnumFacing.NORTH;
        }
    }

    @Override
    public Vec3d getPos() {
        return pos;
    }

    @Nonnull
    @Override
    public List<ItemStack> computeRequiredItems() {
        Set<JsonRule> rules = RulesLoader.getRules(
            new ResourceLocation(entityNbt.func_74779_i("id")),
            entityNbt
        );
        if (rules.isEmpty()) {
            throw new IllegalArgumentException("Rules are empty");
        }
        return rules.stream()
            .map(rule -> rule.requiredExtractors)
            .filter(Objects::nonNull)
            .flatMap(Collection::stream)
            .flatMap(requiredExtractor -> requiredExtractor.extractItemsFromEntity(entityNbt).stream())
            .filter(((Predicate<ItemStack>) ItemStack::func_190926_b).negate())
            .collect(Collectors.toList());
    }

    @Nonnull
    @Override
    public List<FluidStack> computeRequiredFluids() {
        Set<JsonRule> rules = RulesLoader.getRules(
            new ResourceLocation(entityNbt.func_74779_i("id")),
            entityNbt
        );
        return rules.stream()
            .map(rule -> rule.requiredExtractors)
            .filter(Objects::nonNull)
            .flatMap(Collection::stream)
            .flatMap(requiredExtractor -> requiredExtractor.extractFluidsFromEntity(entityNbt).stream())
            .filter(Objects::nonNull)
            .collect(Collectors.toList());
    }

    @Override
    public SchematicEntityDefault getRotated(Rotation rotation) {
        SchematicEntityDefault schematicEntity = SchematicEntityManager.createCleanCopy(this);
        schematicEntity.entityNbt = entityNbt;
        schematicEntity.pos = RotationUtil.rotateVec3d(pos, rotation);
        schematicEntity.hangingPos = hangingPos.func_190942_a(rotation);
        schematicEntity.hangingFacing = rotation.func_185831_a(hangingFacing);
        schematicEntity.entityRotation = entityRotation.func_185830_a(rotation);
        return schematicEntity;
    }

    @Override
    public Entity build(World world, BlockPos basePos) {
        Set<JsonRule> rules = RulesLoader.getRules(
            new ResourceLocation(entityNbt.func_74779_i("id")),
            entityNbt
        );
        NBTTagCompound replaceNbt = rules.stream()
            .map(rule -> rule.replaceNbt)
            .filter(Objects::nonNull)
            .map(NBTBase.class::cast)
            .reduce(NBTUtilBC::merge)
            .map(NBTTagCompound.class::cast)
            .orElse(null);
        Vec3d placePos = new Vec3d(basePos).func_178787_e(pos);
        BlockPos placeHangingPos = basePos.func_177971_a(hangingPos);
        NBTTagCompound newEntityNbt = new NBTTagCompound();
        entityNbt.func_150296_c().stream()
            .map(key -> Pair.of(key, entityNbt.func_74781_a(key)))
            .forEach(kv -> newEntityNbt.func_74782_a(kv.getKey(), kv.getValue()));
        newEntityNbt.func_74782_a("Pos", NBTUtilBC.writeVec3d(placePos));
        newEntityNbt.func_186854_a("UUID", UUID.randomUUID());
        boolean rotate = false;
        if (Stream.of("TileX", "TileY", "TileZ", "Facing").allMatch(newEntityNbt::func_74764_b)) {
            newEntityNbt.func_74768_a("TileX", placeHangingPos.func_177958_n());
            newEntityNbt.func_74768_a("TileY", placeHangingPos.func_177956_o());
            newEntityNbt.func_74768_a("TileZ", placeHangingPos.func_177952_p());
            newEntityNbt.func_74774_a("Facing", (byte) hangingFacing.func_176736_b());
        } else {
            rotate = true;
        }
        Entity entity = EntityList.func_75615_a(
            replaceNbt != null
                ? (NBTTagCompound) NBTUtilBC.merge(newEntityNbt, replaceNbt)
                : newEntityNbt,
            world
        );
        if (entity != null) {
            if (rotate) {
                entity.func_70012_b(
                    placePos.field_72450_a,
                    placePos.field_72448_b,
                    placePos.field_72449_c,
                    entity.field_70177_z + (entity.field_70177_z - entity.func_184229_a(entityRotation)),
                    entity.field_70125_A
                );
            }
            world.func_72838_d(entity);
        }
        return entity;
    }

    @Override
    public Entity buildWithoutChecks(World world, BlockPos basePos) {
        return build(world, basePos);
    }

    @Override
    public NBTTagCompound serializeNBT() {
        NBTTagCompound nbt = new NBTTagCompound();
        nbt.func_74782_a("entityNbt", entityNbt);
        nbt.func_74782_a("pos", NBTUtilBC.writeVec3d(pos));
        nbt.func_74782_a("hangingPos", NBTUtil.func_186859_a(hangingPos));
        nbt.func_74782_a("hangingFacing", NBTUtilBC.writeEnum(hangingFacing));
        nbt.func_74782_a("entityRotation", NBTUtilBC.writeEnum(entityRotation));
        return nbt;
    }

    @Override
    public void deserializeNBT(NBTTagCompound nbt) throws InvalidInputDataException {
        entityNbt = nbt.func_74775_l("entityNbt");
        pos = NBTUtilBC.readVec3d(nbt.func_74781_a("pos"));
        hangingPos = NBTUtil.func_186861_c(nbt.func_74775_l("hangingPos"));
        hangingFacing = NBTUtilBC.readEnum(nbt.func_74781_a("hangingFacing"), EnumFacing.class);
        entityRotation = NBTUtilBC.readEnum(nbt.func_74781_a("entityRotation"), Rotation.class);
    }

    @Override
    public boolean equals(Object o) {
        if (this == o) {
            return true;
        }
        if (o == null || getClass() != o.getClass()) {
            return false;
        }

        SchematicEntityDefault that = (SchematicEntityDefault) o;

        return entityNbt.equals(that.entityNbt) &&
            pos.equals(that.pos) &&
            hangingPos.equals(that.hangingPos) &&
            hangingFacing == that.hangingFacing &&
            entityRotation == that.entityRotation;
    }

    @Override
    public int hashCode() {
        int result = entityNbt.hashCode();
        result = 31 * result + pos.hashCode();
        result = 31 * result + hangingPos.hashCode();
        result = 31 * result + hangingFacing.hashCode();
        result = 31 * result + entityRotation.hashCode();
        return result;
    }
}
